/*	$Id: worker.cpp 3090 2006-01-10 17:07:17Z jgressma $
 *
 *  Copyright 2005 University of Potsdam, Germany
 * 
 *	This file is part of Platypus.
 *
 *  Platypus is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Platypus is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Platypus; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <sys/lwp.h>
#include <vector>
#include <unistd.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include <sys/types.h>
#include <cassert>
#include <platypus/exceptions.h>
#include <platypus/cores/kernel.h>
#include <platypus/cores/foreman.h>
#include <platypus/cores/worker.h>

namespace Platypus
{
	/************************************************************************/
	/* WorkerCallbackAdaptor                                                    */
	/************************************************************************/
	class WorkerCallbackAdaptor	: public KernelCallback
	{
	public:
		WorkerCallbackAdaptor(int id, Foreman& df, ForemanAssistant& tf);
	private: // KernelCallback
		bool shutdown();
		bool delegate();		
		void delegate(const DelegatableChoice& dc);
		bool branch();
		void branch(const DelegatableChoice& dc);
		Expander* expander(const DelegatableChoice& dc = DelegatableChoice()); 
		DelegatableChoicePolicy::PolicyBase* delegatableChoicePolicy();
		DelegatableChoicePolicy::PolicyBase* threadDelegationPolicy();
		ProbingPolicy::PolicyBase* probingPolicy();
		void answerSet(const PartialAssignment& pa);
		void markConflict();
		void markAnswerSet();
		void markBacktrack();
		void markExpanderInitialization();
		int processId() const;
		//int threadId() const;
		bool silent() const;
		ProbingManager* probingManager();
		NopStream* output();
		size_t atoms() const;
	public: // methods for Worker
		void exception(const PlatypusException& e);
		DelegatableChoice retrieve();
		bool wait();
		int threadId() const;
	private:
		const int id_;
		Foreman* foreman_;
		ForemanAssistant* assistant_;
	};

	WorkerCallbackAdaptor::WorkerCallbackAdaptor(int id, Foreman& df, ForemanAssistant& tf)
		:	id_(id)
		,	foreman_(&df)
		,	assistant_(&tf)
	{}
	
	bool WorkerCallbackAdaptor::delegate()
	{
		return foreman_->delegate();
	}
	bool WorkerCallbackAdaptor::shutdown()
	{
		return foreman_->shutdown();
	}
	void WorkerCallbackAdaptor::delegate(const DelegatableChoice& dc)
	{
		foreman_->delegate(dc);
	}
	Expander* WorkerCallbackAdaptor::expander(const DelegatableChoice& dc)
	{
		return assistant_->expander(dc);
	}
	
	DelegatableChoicePolicy::PolicyBase* WorkerCallbackAdaptor::delegatableChoicePolicy()
	{
		return assistant_->delegatableChoicePolicy();
	}
	DelegatableChoicePolicy::PolicyBase* WorkerCallbackAdaptor::threadDelegationPolicy()
	{
		return assistant_->threadDelegationPolicy();
	}
	ProbingPolicy::PolicyBase* WorkerCallbackAdaptor::probingPolicy()
	{
		return assistant_->probingPolicy();
	}
	void WorkerCallbackAdaptor::answerSet(const PartialAssignment& pa)
	{
		foreman_->answerSet(pa);
	}
	void WorkerCallbackAdaptor::markConflict()
	{
		foreman_->markConflict();
	}
	void WorkerCallbackAdaptor::markAnswerSet()
	{
		foreman_->markAnswerSet();
	}
	void WorkerCallbackAdaptor::markBacktrack()
	{
		foreman_->markBacktrack();
	}
	void WorkerCallbackAdaptor::markExpanderInitialization()
	{
		foreman_->markExpanderInitialization();
	}
	int WorkerCallbackAdaptor::processId() const
	{
		return assistant_->processId();
	}
	int WorkerCallbackAdaptor::threadId() const
	{
		return id_;
	}
	bool WorkerCallbackAdaptor::silent() const
	{
		return assistant_->silent();
	}
	ProbingManager* WorkerCallbackAdaptor::probingManager()
	{
		return assistant_->probingManager();
	}
	NopStream* WorkerCallbackAdaptor::output()
	{
		return assistant_->output();
	}
	size_t WorkerCallbackAdaptor::atoms() const
	{
		return assistant_->atoms();
	}
	bool WorkerCallbackAdaptor::branch()
	{
		return foreman_->branch();
	}
	void WorkerCallbackAdaptor::branch(const DelegatableChoice& dc)
	{
		foreman_->branch(dc);
	}
	void WorkerCallbackAdaptor::exception(const PlatypusException& e)
	{
		foreman_->exception(e);
	}
	DelegatableChoice WorkerCallbackAdaptor::retrieve()
	{
		return foreman_->retrieve();
	}
	bool WorkerCallbackAdaptor::wait()
	{
		return foreman_->wait();
	}

	/************************************************************************/
	/* Worker                                                               */
	/************************************************************************/
	Worker::~Worker()
	{}
	Worker::Worker(int id, Foreman& df, ForemanAssistant& tf)
		:	callback_(new WorkerCallbackAdaptor(id, df, tf))
		,	kernel_(new Kernel(callback_.get()))
	{}
	void Worker::threadMain()
	{
		try
		{
                      processorid_t cpuid_max = sysconf(_SC_CPUID_MAX);
                      std::vector<processorid_t> processorids;
                      processorid_t online = 0;
                      for (int i = 0; i <= cpuid_max; i++) {
                
                            // put the id into the iterator if online
        
                            if (p_online(i, P_STATUS) != -1) {
                              online++;
                              processorids.push_back(i);
        
                              }
                      }
        
                      processorid_t pid = processorids.at(callback_->threadId() % online);   
                      // bind the thread to processor
                      lwpid_t myid = _lwp_self();
                      processor_bind(P_LWPID,myid,pid,NULL);

			kernel_->configure();
			while(!callback_->wait())
			{
				const DelegatableChoice dc = callback_->retrieve();
				kernel_->run(dc);
			}
		}
		catch(const PlatypusException& e)
		{
			callback_->exception(e);
		}
		catch(...)
		{
			const PlatypusException e("Non-Platypus exception in thread");
			callback_->exception(e);
		}
	}
}
